#include <errno.h>
#include <math.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include "configuration.h"

static bool get_int(char** s, void* p) {
   int* ip = (int*) p;
   char* endptr;
   errno = 0;
   long int val = strtol(*s, &endptr, 10);
   if (errno || val <= 0) return false;
   *ip = val;
   *s = endptr;
   return true;
}

static bool get_double(char** s, void* p) {
   double* dp = (double*) p;
   char* endptr;
   errno = 0;
   double val = strtod(*s, &endptr);
   if (errno) return false;
   *dp = val;
   *s = endptr;
   return true;
}

static bool skip_colon(char** s) {
   if (**s != ':') {
      printf("colon expected at '%s'\n", *s); // salami
      return false;
   }
   ++*s;
   return true;
}

bool get_magnet(char* s, Magnet* magnet) {
   return
      get_double(&s, &magnet->pos[0]) && skip_colon(&s) &&
      get_double(&s, &magnet->pos[1]) && skip_colon(&s) &&
      get_double(&s, &magnet->strength) && skip_colon(&s) &&
      get_double(&s, &magnet->size) && skip_colon(&s) &&
      get_double(&s, &magnet->color.red) && skip_colon(&s) &&
      get_double(&s, &magnet->color.green) && skip_colon(&s) &&
      get_double(&s, &magnet->color.blue) && !*s;
}

bool get_magnet_with_polar_coordinates(char* s, Magnet* magnet) {
   double theta; double r;
   bool ok =
      get_double(&s, &theta) && skip_colon(&s) &&
      get_double(&s, &r) && skip_colon(&s) &&
      get_double(&s, &magnet->strength) && skip_colon(&s) &&
      get_double(&s, &magnet->size) && skip_colon(&s) &&
      get_double(&s, &magnet->color.red) && skip_colon(&s) &&
      get_double(&s, &magnet->color.green) && skip_colon(&s) &&
      get_double(&s, &magnet->color.blue) && !*s;
   if (ok) {
      theta *= 2 * M_PI / 360;
      magnet->pos[0] = r * cos(theta);
      magnet->pos[1] = r * sin(theta);
   }
   return ok;
}

static Configuration c;

#define OFFSET(base, member) \
   ((char*)&(base.member) - (char*)&(base))

#define MEMBER(member) OFFSET(c, member)

static struct Option {
   char name;
   const char* description;
   bool (*input)(char** s, void* p);
   size_t displacement;
} options[] = {
   {'f', "friction", get_double, MEMBER(friction)},
   {'d', "delta t", get_double, MEMBER(deltat)},
   {'H', "Hooke factor", get_double, MEMBER(hooke_factor)},
   {'P', "height of pendulum", get_double, MEMBER(pendulum_height)},
   {'t', "velocity threshold", get_double, MEMBER(threshold)},
   {'M', "maximal number of iterations", get_int, MEMBER(max_iterations)},
   {'x', "image centre, x coordinate", get_double, MEMBER(center_x)},
   {'y', "image centre, y coordinate", get_double, MEMBER(center_y)},
   {'D', "horizontal diameter", get_double, MEMBER(horizontal_diameter)},
   {'w', "image width", get_int, MEMBER(width)},
   {'h', "image height", get_int, MEMBER(height)},
};
typedef struct Option Option;
const unsigned int nofoptions = sizeof(options)/sizeof(options[0]);

void usage(const char* cmdname) {
   fprintf(stderr, "Usage: %s", cmdname);
   for (int i = 0; i < nofoptions; ++i) {
      fprintf(stderr, " [-%c value]", options[i].name);
   }
   fprintf(stderr, " {-m x:y:strength:size:red:green:blue}");
   fprintf(stderr, " ...\n");
   for (int i = 0; i < nofoptions; ++i) {
      fprintf(stderr, "   -%c value: %s\n",
	 options[i].name, options[i].description);
   }
}

#define CLEANUP (conf? (free(conf->magnets), free(conf), 0): 0)

#define FATAL(arg, msg) \
   (CLEANUP, \
   fprintf(stderr, "%s: at argument \"%s\": %s\n", (cmdname), (arg), (msg)), \
   usage(cmdname))

Configuration* get_conf_from_args(char* cmdname, int* argc_p, char*** argv_p) {
   Configuration* conf = (Configuration*) malloc(sizeof(Configuration));
   if (!conf) {
      FATAL("", "out of space"); return 0;
   }
   *conf = (Configuration){
      .magnets = 0,
      .number_of_magnets = 0,
      .friction = 0.001,
      .deltat = 2,
      .hooke_factor = 0.000005,
      .pendulum_height = 10,
      .threshold = 0.5,
      .center_x = 0, .center_y = 0,
      .max_iterations = 10000,
      .horizontal_diameter = 400,
      .width = 300, .height = 300,
   };
   unsigned int allocated = 0; /* size of allocated magnets array */
   int argc = *argc_p; char** argv = *argv_p;
   while (argc > 0 && argv[0][0] == '-') {
      if (argv[0][1] == '-' && argv[0][2] == 0) {
	 --argc; ++argv; break;
      }
      char option_ch = argv[0][1];
      int i;
      for (i = 0; i < nofoptions; ++i) {
	 if (options[i].name == option_ch) break;
      }
      if (i < nofoptions) {
	 Option* option = &options[i];
	 char* args;
	 if (argv[0][2]) {
	    args = argv[0] + 2;
	 } else {
	    --argc; ++argv;
	    if (argc == 0) {
	       FATAL(argv[-1], "argument missing"); return 0;
	    }
	    args = argv[0];
	 }
	 void* valptr = (void*) ((char*) conf + option->displacement);
	 if (!option->input(&args, valptr) || *args) {
	    FATAL(argv[0], "invalid argument"); return 0;
	 }
      } else {
	 switch (option_ch) {
	    case 'm':
	    case 'p':
	       {
		  char* args;
		  if (argv[0][2]) {
		     args = argv[0] + 2;
		  } else {
		     --argc; ++argv;
		     if (argc == 0) {
			FATAL(argv[-1], "argument missing"); return 0;
		     }
		     args = argv[0];
		  }
		  if (conf->number_of_magnets == allocated) {
		     allocated = allocated * 2 + 3;
		     Magnet* magnets =
			realloc(conf->magnets, sizeof(Magnet) * allocated);
		     if (magnets == 0) {
			FATAL(argv[0], "out of memory"); return 0;
		     }
		     conf->magnets = magnets;
		  }
		  bool ok;
		  if (option_ch == 'm') {
		     ok = get_magnet(args,
			   &conf->magnets[conf->number_of_magnets]);
		  } else {
		     ok = get_magnet_with_polar_coordinates(args,
			   &conf->magnets[conf->number_of_magnets]);
		  }
		  if (!ok) {
		     FATAL(argv[0], "invalid magnet specification"); return 0;
		  }
		  ++conf->number_of_magnets;
	       }
	       break;
	    default:
	       FATAL(argv[0], "unknown option"); return 0;
	 }
      }
      --argc; ++argv;
   }
   *argc_p = argc; *argv_p = argv;
   return conf;
}
