#include <math.h>
#include "configuration.h"
#include "simulation.h"

static inline double sqr(double x) {
   return x * x;
}

static inline double abs1(double vec[2]) {
   return sqrt(sqr(vec[0]) + sqr(vec[1]));
}

static inline double abs2(double vec1[2], double vec2[2]) {
   return sqrt(sqr(vec1[0] - vec2[0]) + sqr(vec1[1] - vec2[1]));
}

SimulationResult simulate(double startx, double starty,
      const Configuration* conf) {
   double pos[2] = {startx, starty};
   double vel[2] = {0, 0};
   double acc[3][2] = {{0, 0}, {0, 0}, {0, 0}};
   unsigned int ai_prev = 0, ai = 1, ai_next = 2;
   double dt = conf->deltat;
   double k = conf->hooke_factor;
   double height = conf->pendulum_height;
   unsigned int iterations = 0;
   double duration = 0;
   double friction = conf->friction;
   double threshold = conf->threshold;

   for(;;) {
      for (unsigned int i = 0; i < 2; ++i) {
	 pos[i] += vel[i]*dt + sqr(dt) *
	    (2/3.0 * acc[ai][i] - 1/6.0 * acc[ai_prev][i]);
      }
      // apply Hooke's law:
      for (unsigned int i = 0; i < 2; ++i) {
	 acc[ai_next][i] = - pos[i] * k;
      }
      // apply forces of the magnets
      for (unsigned int mi = 0; mi < conf->number_of_magnets; ++mi) {
	 Magnet* magnet = conf->magnets + mi;
	 double dist =
	    sqrt(
	       sqr(magnet->pos[0] - pos[0]) +
	       sqr(magnet->pos[1] - pos[1]) +
	       sqr(height)
	    );
	 double dist3 = dist * dist * dist;
	 for (unsigned int i = 0; i < 2; ++i) {
	    acc[ai_next][i] -= magnet->strength / dist3 *
	       (pos[i] - magnet->pos[i]);
	 }
      }
      // apply friction
      for (unsigned int i = 0; i < 2; ++i) {
	 acc[ai_next][i] -= vel[i] * friction;
      }
      // Beeman integration for velocities
      for (unsigned int i = 0; i < 2; ++i) {
	 vel[i] += dt * (1/3.0 * acc[ai_next][i] +
	    5/6.0 * acc[ai][i] - 1/6.0 * acc[ai_prev][i]);
      }

      // advance to next iteration
      ai_prev = (ai_prev + 1) % 2;
      ai = (ai + 1) % 2;
      ai_next = (ai_next + 1) % 2;
      ++iterations; duration += dt;

      for (unsigned int mi = 0; mi < conf->number_of_magnets; ++mi) {
	 Magnet* magnet = conf->magnets + mi;
	 // check if we our simulation ends at this magnet
	 if (abs1(vel) < threshold && abs2(magnet->pos, pos) < magnet->size) {
	    return (SimulationResult){
	       .magnet_index = mi,
	       .duration = duration,
	       .iterations = iterations,
	       .success = true,
	    };
	 }
      }
      if (iterations == conf->max_iterations) {
	 return (SimulationResult){.iterations = iterations, .success = false};
      }
   }
}
